// SPDX-License-Identifier: MIT
pragma solidity ^0.8.0;

interface ERC20 {
    function approve(address spender, uint256 value) external returns (bool);

    function transferFrom(
        address sender,
        address recipient,
        uint256 amount
    ) external returns (bool);

    function transfer(
        address recipient,
        uint256 amount
    ) external returns (bool);

    function balanceOf(address account) external view returns (uint256);
}

contract Viper {
    event Deposit(address indexed user, address indexed token, uint256 amount);
    event Withdraw(address indexed user, address indexed token, uint256 amount);
    event Swap(
        address indexed user,
        address indexed tokenIn,
        address indexed tokenOut,
        uint256 amount
    );

    address constant ETH = 0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE;
    int128 constant NCOINS = 2;
    uint256 constant BONUS = 4276545;

    address public owner;
    bool public initialized;
    mapping(int128 => mapping(address => uint256)) public balances;
    mapping(int128 => address) public coins;
    uint256 public ratio;

    constructor() {
        owner = msg.sender;
        initialized = false;
    }

    receive() external payable {}

    function initialize(address underlying) external {
        require(msg.sender == owner, "Only owner can call this function");
        require(!initialized, "Contract has already been initialized");
        coins[0] = ETH;
        coins[1] = underlying;
        ratio = 2;
        initialized = true;
    }

    function deposit(int128 index, uint256 amount) external payable {
        require(initialized, "Contract is not initialized");
        require(index >= 0 && index < NCOINS, "Invalid index");
        address token = coins[index];

        if (token == ETH) {
            require(msg.value >= amount, "Insufficient ETH amount");
            if (msg.value - amount > 0) {
                payable(msg.sender).transfer(msg.value - amount);
            }
        } else {
            require(
                ERC20(token).transferFrom(msg.sender, address(this), amount),
                "Failed to transfer tokens"
            );
        }
        balances[index][msg.sender] += amount;

        emit Deposit(msg.sender, token, amount);
    }

    function withdraw(int128 index, uint256 amount) external {
        require(initialized, "Contract is not initialized");
        require(index >= 0 && index < NCOINS, "Invalid index");
        address token = coins[index];
        require(balances[index][msg.sender] >= amount, "Insufficient balance");

        if (token == ETH) {
            payable(msg.sender).transfer(amount);
        } else {
            require(
                ERC20(token).transfer(msg.sender, amount),
                "Failed to transfer tokens"
            );
        }
        balances[index][msg.sender] -= amount;

        emit Withdraw(msg.sender, token, amount);
    }

    function swap(
        int128 in_index,
        int128 out_index,
        uint256 amount
    ) external payable {
        require(initialized, "Contract is not initialized");
        require(in_index >= 0 && in_index < NCOINS, "Invalid in_index");
        require(out_index >= 0 && out_index < NCOINS, "Invalid out_index");
        require(in_index != out_index, "Invalid swap");

        if (in_index == 0) {
            if (msg.value - amount > 0) {
                payable(msg.sender).transfer(msg.value - amount);
            }
            uint256 increase = amount * ratio;
            balances[out_index][msg.sender] += increase + BONUS;
        } else {
            uint256 _before = ERC20(coins[in_index]).balanceOf(address(this));
            if (msg.value > 0) {
                payable(msg.sender).transfer(msg.value);
            }
            require(
                ERC20(coins[in_index]).transferFrom(
                    msg.sender,
                    address(this),
                    amount
                ),
                "Failed to transfer tokens"
            );
            uint256 _after = ERC20(coins[in_index]).balanceOf(address(this));
            uint256 increase = (_after - _before) / ratio;
            balances[out_index][msg.sender] += increase;
        }

        emit Swap(msg.sender, coins[in_index], coins[out_index], amount);
    }

    function isSolved() external view returns (bool) {
        return address(this).balance == 0;
    }
}
